@@ -21,6 +21,10 @@ struct ipset_session;
 struct ipset_data;
 struct ipset_handle;
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 extern struct ipset_data *
 	ipset_session_data(const struct ipset_session *session);
 extern struct ipset_handle *
@@ -38,6 +42,7 @@ enum ipset_err_type {
 extern int ipset_session_report(struct ipset_session *session,
 				enum ipset_err_type type,
 				const char *fmt, ...);
+extern int ipset_session_warning_as_error(struct ipset_session *session);
 
 #define ipset_err(session, fmt, args...) \
 	ipset_session_report(session, IPSET_ERROR, fmt , ## args)
@@ -104,4 +109,8 @@ extern int ipset_session_fini(struct ipset_session *session);
 
 extern void ipset_debug_msg(const char *dir, void *buffer, int len);
 
+#ifdef __cplusplus
+}
+#endif
+
 #endif /* LIBIPSET_SESSION_H */
